;;; Directory Local Variables
;;; For more information see (info "(emacs) Directory Variables")

((emacs-lisp-mode . ((sentence-end-double-space . nil)
                     (fill-column . 110)
                     (indent-tabs-mode . nil)
                     (elisp-lint-indent-specs . ((describe . 1)
                                                 (it . 1)
                                                 (thread-first . 0)
                                                 (cl-flet . 1)
                                                 (cl-flet* . 1))))))
